/**********************************************/
/* CPU_Disassembler.cpp                       */
/*--------------------------------------------*/
/* POTATO Chip Instruction Set Disassembler   */
/*--------------------------------------------*/
/* Copyright (C) 2002 Michel Gerritse         */
/*                                            */
/* Last update: 16-10-2002                    */
/*--------------------------------------------*/
/* TODO:                                      */
/**********************************************/
#include "VMUCore.h"
#include <string.h>
#include <stdio.h>

char Label[0x20];

char* GetLabel(unsigned short Addr)
{
  if (Addr < 0x100)
  {
    sprintf(Label, "RAM0x%02x", Addr);
  }
  else if (Addr >= 0x180)
  {
    sprintf(Label, "XRAM0x%03x", Addr - 0x180);
  }
  else
  {
    switch (Addr)
    {
      case 0x100:
        sprintf(Label, "Acc");
        break;

      case 0x101:
        sprintf(Label, "PSW");
        break;

      case 0x102:
        sprintf(Label, "B");
        break;

      case 0x103:
        sprintf(Label, "C");
        break;

      case 0x104:
        sprintf(Label, "TRL");
        break;

      case 0x105:
        sprintf(Label, "TRH");
        break;

      case 0x106:
        sprintf(Label, "SP");
        break;

      case 0x107:
        sprintf(Label, "PCON");
        break;

      case 0x108:
        sprintf(Label, "IE");
        break;

      case 0x109:
        sprintf(Label, "IP");
        break;

      case 0x10D:
        sprintf(Label, "EXT");
        break;

      case 0x10E:
        sprintf(Label, "OCR");
        break;

      case 0x110:
        sprintf(Label, "T0CNT");
        break;

      case 0x111:
        sprintf(Label, "T0PRR");
        break;

      case 0x112:
        sprintf(Label, "T0L");
        break;

      case 0x113:
        sprintf(Label, "T0LR");
        break;

      case 0x114:
        sprintf(Label, "T0H");
        break;

      case 0x115:
        sprintf(Label, "T0HR");
        break;

      case 0x118:
        sprintf(Label, "T1CNT");
        break;

      case 0x11A:
        sprintf(Label, "T1LC");
        break;

      case 0x11B:
        sprintf(Label, "T1L");
        break;

      case 0x11C:
        sprintf(Label, "T1LR/T1HC");
        break;

      case 0x11D:
        sprintf(Label, "T1H/T1HR");
        break;

      case 0x120:
        sprintf(Label, "MCR");
        break;

      case 0x122:
        sprintf(Label, "STAD");
        break;

      case 0x123:
        sprintf(Label, "CNR");
        break;

      case 0x124:
        sprintf(Label, "TDR");
        break;

      case 0x125:
        sprintf(Label, "XBNK");
        break;

      case 0x127:
        sprintf(Label, "VCCR");
        break;

      case 0x144:
        sprintf(Label, "P1");
        break;

      case 0x145:
        sprintf(Label, "P1DDR");
        break;

      case 0x146:
        sprintf(Label, "P1FCR");
        break;

      case 0x14C:
        sprintf(Label, "P3");
        break;

      case 0x14D:
        sprintf(Label, "P3DDR");
        break;

      case 0x14E:
        sprintf(Label, "P3INT");
        break;

      case 0x15C:
        sprintf(Label, "P7");
        break;

      case 0x15D:
        sprintf(Label, "I01CR");
        break;

      case 0x15E:
        sprintf(Label, "I23CR");
        break;

      case 0x15F:
        sprintf(Label, "ISL");
        break;

      case 0x163:
        sprintf(Label, "VSEL");
        break;

      case 0x164:
        sprintf(Label, "VRMAD1");
        break;

      case 0x165:
        sprintf(Label, "VRMAD2");
        break;

      case 0x166:
        sprintf(Label, "VTRBF");
        break;

      case 0x17F:
        sprintf(Label, "BTCR");
        break;

      default:
        sprintf(Label, "SFR0x%03x",Addr);
        break;
    }
  }

  return Label;
}


/*------------------------------------*/
/*--   Disassemble 1 Instruction    --*/
/*------------------------------------*/
/* Input : ProgramCounter, TextBuffer */
/* Output: Nr of Bytes Opcode uses    */
/* Action: Read opcode at specified   */
/*         PC, and disassemble it.    */
/*------------------------------------*/
/* Notes : You can get the next PC    */
/*         by adding the return       */
/*         value to the current PC    */
/*------------------------------------*/
int POTATO_Disassemble(unsigned short PC, char *TextBuffer)
{
  char           temp[0x100];
  unsigned short Data;
  unsigned short OldPC  = PC;
  unsigned char  Opcode = Cpu.Code[PC];
  int            Bytes  = 1;

  switch (Opcode >> 4)
  {
    case 0x0:
      switch (Opcode & 0xF)
      {
        case 0x0:
          sprintf(temp, "NOP");
          Bytes = 1;
          break;

        case 0x1:
          sprintf(temp, "BR 0x%04x", PC + 2 + (char)(Cpu.Code[PC + 1]));
          Bytes = 2;
          break;

        case 0x2:
        case 0x3:
          sprintf(temp, "LD %s", GetLabel(((Opcode & 1) << 8) | Cpu.Code[PC + 1]));
          Bytes = 2;
          break;

        case 0x4:
        case 0x5:
        case 0x6:
        case 0x7:
          sprintf(temp, "LD @R%d", Opcode & 3);
          Bytes = 1;
          break;

        case 0x8:
        case 0x9:
        case 0xA:
        case 0xB:
        case 0xC:
        case 0xD:
        case 0xE:
        case 0xF:
          Data = (((Opcode & 0x10) >> 1) | (Opcode & 0x7)) << 8;
          Data |= Cpu.Code[PC + 1];
          sprintf(temp, "CALL 0x%04x", (PC & 0xF000) | (Data & 0x0FFF));
          Bytes = 2;
          break;
      }
      break;
    
    case 0x1:
      switch (Opcode & 0xF)
      {
        case 0x0:
          sprintf(temp, "CALLR 0x%0x4", PC + 2 + (Cpu.Code[PC + 2] << 8) | Cpu.Code[PC + 1]);
          Bytes = 3;
          break;

        case 0x1:
          sprintf(temp, "BRF 0x%04x", PC + 2 + ((Cpu.Code[PC + 2] << 8) | Cpu.Code[PC + 1]));
          Bytes = 3;
          break;

        case 0x2:
        case 0x3:
          sprintf(temp, "ST %s",  GetLabel(((Opcode & 1) << 8) | Cpu.Code[PC + 1]));
          Bytes = 2;
          break;

        case 0x4:
        case 0x5:
        case 0x6:
        case 0x7:
          sprintf(temp, "ST @R%d", Opcode & 3);
          Bytes = 1;
          break;

        case 0x8:
        case 0x9:
        case 0xA:
        case 0xB:
        case 0xC:
        case 0xD:
        case 0xE:
        case 0xF:
          Data = (((Opcode & 0x10) >> 1) | (Opcode & 0x7)) << 8;
          Data |= Cpu.Code[PC + 1];
          sprintf(temp, "CALL 0x%04x", (PC & 0xF000) | (Data & 0x0FFF));
          Bytes = 2;
          break;
      }
      break;

    case 0x2:
      switch (Opcode & 0xF)
      {
        case 0x0:
          sprintf(temp, "CALLF 0x%04x", (Cpu.Code[PC + 1] << 8) | Cpu.Code[PC + 2]);
          Bytes = 3;
          break;

        case 0x1:
          sprintf(temp, "JUMPF 0x%04x", (Cpu.Code[PC + 1] << 8) | Cpu.Code[PC + 2]);
          Bytes = 3;
          break;

        case 0x2:
        case 0x3:
          sprintf(temp, "MOV #0x%02x, %s", Cpu.Code[PC + 2], GetLabel(((Opcode & 1) << 8) | Cpu.Code[PC + 1]));
          Bytes = 3;
          break;

        case 0x4:
        case 0x5:
        case 0x6:
        case 0x7:
          sprintf(temp, "MOV #0x%02x, @R%d", Cpu.Code[PC + 1], Opcode & 3);
          Bytes = 2;
          break;

        case 0x8:
        case 0x9:
        case 0xA:
        case 0xB:
        case 0xC:
        case 0xD:
        case 0xE:
        case 0xF:
          Data = (((Opcode & 0x10) >> 1) | (Opcode & 0x7)) << 8;
          Data |= Cpu.Code[PC + 1];
          PC += 2;
          sprintf(temp, "JMP 0x%04x", (PC & 0xF000) | (Data & 0x0FFF));
          Bytes = 2;
          break;
      }
      break;

    case 0x3:
      switch (Opcode & 0xF)
      {
        case 0x0:
          sprintf(temp, "MUL");
          Bytes = 1;
          break;

        case 0x1:
          sprintf(temp, "BE #0x%02x, 0x%04x", Cpu.Code[PC + 1], PC + 3 + (char)Cpu.Code[PC + 2]);
          Bytes = 3;
          break;

        case 0x2:
        case 0x3:
          sprintf(temp, "BE %s, 0x%04x", GetLabel(((Opcode & 1) << 8) | Cpu.Code[PC + 1]), PC + 3 + (char)Cpu.Code[PC + 2]);
          Bytes = 3;
          break;

        case 0x4:
        case 0x5:
        case 0x6:
        case 0x7:
          sprintf(temp, "BE @R%d, #0x%02x, 0x%04x", Opcode & 3, Cpu.Code[PC + 1], PC + 3 + (char)Cpu.Code[PC + 2]);
          Bytes = 3;
          break;

        case 0x8:
        case 0x9:
        case 0xA:
        case 0xB:
        case 0xC:
        case 0xD:
        case 0xE:
        case 0xF:
          Data = (((Opcode & 0x10) >> 1) | (Opcode & 0x7)) << 8;
          Data |= Cpu.Code[PC + 1];
          PC += 2;
          sprintf(temp, "JMP 0x%04x", (PC & 0xF000) | (Data & 0x0FFF));
          Bytes = 2;
          break;
      }
      break;

    case 0x4:
      switch (Opcode & 0xF)
      {
        case 0x0:
          sprintf(temp, "DIV");
          Bytes = 1;
          break;

        case 0x1:
          sprintf(temp, "BNE #0x%02x, 0x%04x", Cpu.Code[PC + 1], PC + 3 + (char)Cpu.Code[PC + 2]);
          Bytes = 3;
          break;

        case 0x2:
        case 0x3:
          sprintf(temp, "BNE %s, 0x%04x", GetLabel(((Opcode & 1) << 8) | Cpu.Code[PC + 1]), PC + 3 + (char)Cpu.Code[PC + 2]);
          Bytes = 3;
          break;

        case 0x4:
        case 0x5:
        case 0x6:
        case 0x7:
          sprintf(temp, "BNE @R%d, #0x%02x, 0x%04x", Opcode & 3, Cpu.Code[PC + 1], PC + 3 + (char)Cpu.Code[PC + 2]);
          Bytes = 3;
          break;

        case 0x8:
        case 0x9:
        case 0xA:
        case 0xB:
        case 0xC:
        case 0xD:
        case 0xE:
        case 0xF:
          sprintf(temp, "BPC %s, %d, 0x%04x", GetLabel(((Opcode & 0x10) << 4) | Cpu.Code[PC + 1]), Opcode & 7, PC + 3 + (char)Cpu.Code[PC + 2]);
          Bytes = 3;
          break;
      }
      break;

    case 0x5:
      switch (Opcode & 0xF)
      {
        case 0x0:
          sprintf(temp, "LDF");
          Bytes = 1;
          break;

        case 0x1:
          sprintf(temp, "STF");
          Bytes = 1;
          break;

        case 0x2:
        case 0x3:
          sprintf(temp, "DBNZ %s, 0x%04x", GetLabel(((Opcode & 0x1) << 8) | Cpu.Code[PC + 1]), PC + 3 + (char)Cpu.Code[PC + 2]);
          Bytes = 3;
          break;

        case 0x4:
        case 0x5:
        case 0x6:
        case 0x7:
          sprintf(temp, "DBNZ @R%d, 0x%04x", Opcode & 3, PC + 2 + (char)Cpu.Code[PC + 2]);
          Bytes = 2;
          break;

        case 0x8:
        case 0x9:
        case 0xA:
        case 0xB:
        case 0xC:
        case 0xD:
        case 0xE:
        case 0xF:
          sprintf(temp, "BPC %s, %d, 0x%04x", GetLabel(((Opcode & 0x10) << 4) | Cpu.Code[PC + 1]), Opcode & 7, PC + 3 + (char)Cpu.Code[PC + 2]);
          Bytes = 3;
          break;
      }
      break;

    case 0x6:
      switch (Opcode & 0xF)
      {
        case 0x0:
        case 0x1:
          sprintf(temp, "PUSH %s", GetLabel(((Opcode & 0x1) << 8) | Cpu.Code[PC + 1]));
          Bytes = 2;
          break;

        case 0x2:
        case 0x3:
          sprintf(temp, "INC %s", GetLabel(((Opcode & 0x1) << 8) | Cpu.Code[PC + 1]));
          Bytes = 2;
          break;

        case 0x4:
        case 0x5:
        case 0x6:
        case 0x7:
          sprintf(temp, "INC @R%d", Opcode & 3);
          Bytes = 1;
          break;

        case 0x8:
        case 0x9:
        case 0xA:
        case 0xB:
        case 0xC:
        case 0xD:
        case 0xE:
        case 0xF:
          sprintf(temp, "BP %s, %d, 0x%04x", GetLabel(((Opcode & 0x10) << 4) | Cpu.Code[PC + 1]), Opcode & 7, PC + 3 + (char)Cpu.Code[PC + 2]);
          Bytes = 3;
          break;
      }
      break;

    case 0x7:
      switch (Opcode & 0xF)
      {
        case 0x0:
        case 0x1:
          sprintf(temp, "POP %s", GetLabel(((Opcode & 0x1) << 8) | Cpu.Code[PC + 1]));
          Bytes = 2;
          break;

        case 0x2:
        case 0x3:
          sprintf(temp, "DEC %s", GetLabel(((Opcode & 0x1) << 8) | Cpu.Code[PC + 1]));
          Bytes = 2;
          break;

        case 0x4:
        case 0x5:
        case 0x6:
        case 0x7:
          sprintf(temp, "DEC @R%d", Opcode & 3);
          Bytes = 1;
          break;

        case 0x8:
        case 0x9:
        case 0xA:
        case 0xB:
        case 0xC:
        case 0xD:
        case 0xE:
        case 0xF:
          sprintf(temp, "BP %s, %d, 0x%04x", GetLabel(((Opcode & 0x10) << 4) | Cpu.Code[PC + 1]), Opcode & 7, PC + 3 + (char)Cpu.Code[PC + 2]);
          Bytes = 3;
          break;
      }
      break;

    case 0x8:
      switch (Opcode & 0xF)
      {
        case 0x0:
          sprintf(temp, "BZ 0x%04x", PC + 2 + (char)Cpu.Code[PC + 1]);
          Bytes = 2;
          break;

        case 0x1:
          sprintf(temp, "ADD #0x%02x", Cpu.Code[PC + 1]);
          Bytes = 2;
          break;

        case 0x2:
        case 0x3:
          sprintf(temp, "ADD %s", GetLabel(((Opcode & 1) << 8) | Cpu.Code[PC + 1]));
          Bytes = 2;
          break;

        case 0x4:
        case 0x5:
        case 0x6:
        case 0x7:
          sprintf(temp, "ADD @R%d", Opcode & 3);
          Bytes = 1;
          break;

        case 0x8:
        case 0x9:
        case 0xA:
        case 0xB:
        case 0xC:
        case 0xD:
        case 0xE:
        case 0xF:
          sprintf(temp, "BN %s, %d, 0x%04x", GetLabel(((Opcode & 0x10) << 4) | Cpu.Code[PC + 1]), Opcode & 7, PC + 3 + (char)Cpu.Code[PC + 2]);
          Bytes = 3;
          break;
      }
      break;

    case 0x9:
      switch (Opcode & 0xF)
      {
        case 0x0:
          sprintf(temp, "BNZ 0x%04x", PC + 2 + (char)Cpu.Code[PC + 1]);
          Bytes = 2;
          break;

        case 0x1:
          sprintf(temp, "ADDC #0x%02x", Cpu.Code[PC + 1]);
          Bytes = 2;
          break;

        case 0x2:
        case 0x3:
          sprintf(temp, "ADDC %s", GetLabel(((Opcode & 1) << 8) | Cpu.Code[PC + 1]));
          Bytes = 2;
          break;

        case 0x4:
        case 0x5:
        case 0x6:
        case 0x7:
          sprintf(temp, "ADDC @R%d", Opcode & 3);
          Bytes = 1;
          break;

        case 0x8:
        case 0x9:
        case 0xA:
        case 0xB:
        case 0xC:
        case 0xD:
        case 0xE:
        case 0xF:
          sprintf(temp, "BN %s, %d, 0x%04x", GetLabel(((Opcode & 0x10) << 4) | Cpu.Code[PC + 1]), Opcode & 7, PC + 3 + (char)Cpu.Code[PC + 2]);
          Bytes = 3;
          break;
      }
      break;

    case 0xA:
      switch (Opcode & 0xF)
      {
        case 0x0:
          sprintf(temp, "RET");
          Bytes = 1;
          break;

        case 0x1:
          sprintf(temp, "SUB #0x%02x", Cpu.Code[PC + 1]);
          Bytes = 2;
          break;

        case 0x2:
        case 0x3:
          sprintf(temp, "SUB %s", GetLabel(((Opcode & 1) << 8) | Cpu.Code[PC + 1]));
          Bytes = 2;
          break;

        case 0x4:
        case 0x5:
        case 0x6:
        case 0x7:
          sprintf(temp, "SUB @R%d", Opcode & 3);
          Bytes = 1;
          break;

        case 0x8:
        case 0x9:
        case 0xA:
        case 0xB:
        case 0xC:
        case 0xD:
        case 0xE:
        case 0xF:
          sprintf(temp, "NOT1 %s, %d", GetLabel(((Opcode & 0x10) << 4) | Cpu.Code[PC + 1]), Opcode & 7);
          Bytes = 2;
          break;
      }
      break;

    case 0xB:
      switch (Opcode & 0xF)
      {
        case 0x0:
          sprintf(temp, "RETI");
          Bytes = 1;
          break;

        case 0x1:
          sprintf(temp, "SUBC #0x%02x", Cpu.Code[PC + 1]);
          Bytes = 2;
          break;

        case 0x2:
        case 0x3:
          sprintf(temp, "SUBC %s", GetLabel(((Opcode & 1) << 8) | Cpu.Code[PC + 1]));
          Bytes = 2;
          break;

        case 0x4:
        case 0x5:
        case 0x6:
        case 0x7:
          sprintf(temp, "SUBC @R%d", Opcode & 3);
          Bytes = 1;
          break;

        case 0x8:
        case 0x9:
        case 0xA:
        case 0xB:
        case 0xC:
        case 0xD:
        case 0xE:
        case 0xF:
          sprintf(temp, "NOT1 %s, %d", GetLabel(((Opcode & 0x10) << 4) | Cpu.Code[PC + 1]), Opcode & 7);
          Bytes = 2;
          break;
      }
      break;

    case 0xC:
      switch (Opcode & 0xF)
      {
        case 0x0:
          sprintf(temp, "ROR");
          Bytes = 1;
          break;

        case 0x1:
          sprintf(temp, "LDC");
          Bytes = 1;
          break;

        case 0x2:
        case 0x3:
          sprintf(temp, "XCH %s", GetLabel(((Opcode & 1) << 8) | Cpu.Code[PC + 1]));
          Bytes = 2;
          break;

        case 0x4:
        case 0x5:
        case 0x6:
        case 0x7:
          sprintf(temp, "XCH @R%d", Opcode & 3);
          Bytes = 1;
          break;

        case 0x8:
        case 0x9:
        case 0xA:
        case 0xB:
        case 0xC:
        case 0xD:
        case 0xE:
        case 0xF:
          sprintf(temp, "CLR1 %s, %d", GetLabel(((Opcode & 0x10) << 4) | Cpu.Code[PC + 1]), Opcode & 7);
          Bytes = 2;
          break;
      }
      break;

    case 0xD:
      switch (Opcode & 0xF)
      {
        case 0x0:
          sprintf(temp, "RORC");
          Bytes = 1;
          break;

        case 0x1:
          sprintf(temp, "OR #0x%02x", Cpu.Code[PC + 1]);
          Bytes = 2;
          break;

        case 0x2:
        case 0x3:
          sprintf(temp, "OR %s", GetLabel(((Opcode & 1) << 8) | Cpu.Code[PC + 1]));
          Bytes = 2;
          break;

        case 0x4:
        case 0x5:
        case 0x6:
        case 0x7:
          sprintf(temp, "OR @R%d", Opcode & 3);
          Bytes = 1;
          break;

        case 0x8:
        case 0x9:
        case 0xA:
        case 0xB:
        case 0xC:
        case 0xD:
        case 0xE:
        case 0xF:
          sprintf(temp, "CLR1 %s, %d", GetLabel(((Opcode & 0x10) << 4) | Cpu.Code[PC + 1]), Opcode & 7);
          Bytes = 2;
          break;
      }
      break;

    case 0xE:
      switch (Opcode & 0xF)
      {
        case 0x0:
          sprintf(temp, "ROL");
          Bytes = 1;
          break;

        case 0x1:
          sprintf(temp, "AND #0x%02x", Cpu.Code[PC + 1]);
          Bytes = 2;
          break;

        case 0x2:
        case 0x3:
          sprintf(temp, "AND %s", GetLabel(((Opcode & 1) << 8) | Cpu.Code[PC + 1]));
          Bytes = 2;
          break;

        case 0x4:
        case 0x5:
        case 0x6:
        case 0x7:
          sprintf(temp, "AND @R%d", Opcode & 3);
          Bytes = 1;
          break;

        case 0x8:
        case 0x9:
        case 0xA:
        case 0xB:
        case 0xC:
        case 0xD:
        case 0xE:
        case 0xF:
          sprintf(temp, "SET1 %s, %d", GetLabel(((Opcode & 0x10) << 4) | Cpu.Code[PC + 1]), Opcode & 7);
          Bytes = 2;
          break;
      }
      break;

    case 0xF:
      switch (Opcode & 0xF)
      {
        case 0x0:
          sprintf(temp, "ROLC");
          Bytes = 1;
          break;

        case 0x1:
          sprintf(temp, "XOR #0x%02x", Cpu.Code[PC + 1]);
          Bytes = 2;
          break;

        case 0x2:
        case 0x3:
          sprintf(temp, "XOR %s", GetLabel(((Opcode & 1) << 8) | Cpu.Code[PC + 1]));
          Bytes = 2;
          break;

        case 0x4:
        case 0x5:
        case 0x6:
        case 0x7:
          sprintf(temp, "XOR @R%d", Opcode & 3);
          Bytes = 1;
          break;

        case 0x8:
        case 0x9:
        case 0xA:
        case 0xB:
        case 0xC:
        case 0xD:
        case 0xE:
        case 0xF:
          sprintf(temp, "SET1 %s, %d", GetLabel(((Opcode & 0x10) << 4) | Cpu.Code[PC + 1]), Opcode & 7);
          Bytes = 2;
          break;
      }
      break;
  }

  if (TextBuffer != NULL)
    sprintf(TextBuffer, "0x%04x :: 0x%02x :: %s", OldPC, Opcode, temp);
  
  return Bytes;
}