/**********************************************/
/* CPU_Interface.cpp                          */
/*--------------------------------------------*/
/* POTATO Chip Interface                      */
/*--------------------------------------------*/
/* Copyright (C) 2002 Michel Gerritse         */
/*                                            */
/* Last update: 25-09-2002                    */
/*--------------------------------------------*/
/* TODO:                                      */
/**********************************************/
#include <string.h>
#include <stdio.h>
#include "VMUCore.h"

_POTATO Cpu;                    /* Our virtual POTATO processor :) */
VMU_Opcode *VMU_OpTable[0x100]; /* Opcode Jump Table */
unsigned char Parity[0x100];    /* Parity Table */

int BiosLoaded = 0;

int IntMask;
int IntReq;
unsigned short IntVector[10] = 
{ 
  0x0003, 
  0x000B, 
  0x0013, 
  0x001B, 
  0x0023, 
  0x002B, 
  0x0033, 
  0x003B, 
  0x0043, 
  0x004B 
};

static struct
{
	int         StartAddr;
	int         EndAddr;
	VMU_Opcode *Opcode;

} _InitOpcode[] = 
{
  { 0x00, 0x00, NOP     },
  { 0x01, 0x01, BR      },
  { 0x02, 0x03, LDd9    },
  { 0x04, 0x07, LDrj    },
  { 0x08, 0x0F, CALL    },
  { 0x10, 0x10, CALLR   },
  { 0x11, 0x11, BRF     },
  { 0x12, 0x13, STd9    },
  { 0x14, 0x17, STrj    },
  { 0x18, 0x1F, CALL    },
  { 0x20, 0x20, CALLF   },
  { 0x21, 0x21, JMPF    },
  { 0x22, 0x23, MOVi8d9 },
  { 0x24, 0x27, MOVi8rj },
  { 0x28, 0x2F, JMP     },
  { 0x30, 0x30, MUL     },
  { 0x31, 0x31, BEi8    },
  { 0x32, 0x33, BEd9    },
  { 0x34, 0x37, BErj    },
  { 0x38, 0x3F, JMP     },
  { 0x40, 0x40, DIV     },
  { 0x41, 0x41, BNEi8   },
  { 0x42, 0x43, BNEd9   },
  { 0x44, 0x47, BNErj   },
  { 0x48, 0x4F, BPC     },
  { 0x50, 0x50, LDF     },
  { 0x51, 0x51, STF     },
  { 0x52, 0x53, DBNZd9  },
  { 0x54, 0x57, DBNZrj  },
  { 0x58, 0x5F, BPC     },
  { 0x60, 0x61, PUSH    },
  { 0x62, 0x63, INCd9   },
  { 0x64, 0x67, INCrj   },
  { 0x68, 0x6F, BP      },
  { 0x70, 0x71, POP     },
  { 0x72, 0x73, DECd9   },
  { 0x74, 0x77, DECrj   },
  { 0x78, 0x7F, BP      },
  { 0x80, 0x80, BZ      },
	{ 0x81, 0x81, ADDi8   },
  { 0x82, 0x83, ADDd9   },
  { 0x84, 0x87, ADDrj   },
  { 0x88, 0x8F, BN      },
  { 0x90, 0x90, BNZ     },
  { 0x91, 0x91, ADDCi8  },
  { 0x92, 0x93, ADDCd9  },
  { 0x94, 0x97, ADDCrj  },
  { 0x98, 0x9F, BN      },
  { 0xA0, 0xA0, RET     },
  { 0xA1, 0xA1, SUBi8   },
  { 0xA2, 0xA3, SUBd9   },
  { 0xA4, 0xA7, SUBrj   },
  { 0xA8, 0xAF, NOT1    },
  { 0xB0, 0xB0, RETI    },
  { 0xB1, 0xB1, SUBCi8  },
  { 0xB2, 0xB3, SUBCd9  },
  { 0xB4, 0xB7, SUBCrj  },
  { 0xB8, 0xBF, NOT1    },
  { 0xC0, 0xC0, ROR     },
  { 0xC1, 0xC1, LDC     },
  { 0xC2, 0xC3, XCHd9   },
  { 0xC4, 0xC7, XCHrj   },
  { 0xC8, 0xCF, CLR1    },
  { 0xD0, 0xD0, RORC    },
  { 0xD1, 0xD1, ORi8    },
  { 0xD2, 0xD3, ORd9    },
  { 0xD4, 0xD7, ORrj    },
  { 0xD8, 0xDF, CLR1    },
  { 0xE0, 0xE0, ROL     },
  { 0xE1, 0xE1, ANDi8   },
  { 0xE2, 0xE3, ANDd9   },
  { 0xE4, 0xE7, ANDrj   },
  { 0xE8, 0xEF, SET1    },
  { 0xF0, 0xF0, ROLC    },
  { 0xF1, 0xF1, XORi8   },
  { 0xF2, 0xF3, XORd9   },
  { 0xF4, 0xF7, XORrj   },
  { 0xF8, 0xFF, SET1    },
  {   -1,   -1, ILLEGAL },
};

/*------------------------------------*/
/*-- Initialize the POTATO CPU core --*/
/*------------------------------------*/
/* Input : None                       */
/* Output: None                       */
/* Action: Builds Opcode Table, and   */
/*         resets the CPU             */
/*------------------------------------*/
/* Notes : Only needs to be called    */
/*         once                       */
/*------------------------------------*/
void POTATO_Init()
{
  static int Init = 0;
  int i;
  int Start, End;

  if (!Init)
  {
    /* Clear Opcode Jump Table */
    for (i = 0; i < 0x100; i++)
    {
      VMU_OpTable[i] = ILLEGAL;
    }

    i = 0;

    /* Build Opcode Jump Table */
    while (_InitOpcode[i].StartAddr != -1)
    {
      Start = _InitOpcode[i].StartAddr;
      End   = _InitOpcode[i].EndAddr;

      for (; Start <= End; Start++)
      {
        VMU_OpTable[Start] = _InitOpcode[i].Opcode;
      }

      i++;
    }

    /* Build Parity Table */
    Parity[0] = 0;
    Parity[1] = 1;
    Parity[2] = 1;
    Parity[3] = 0;

    for (i = 0x4; i < 0x10; i++)
    {
      Parity[i] = Parity[i & 0x3] ^ Parity[i >> 2];
    }

    for (i = 0x10; i < 0x100; i++)
    {
      Parity[i] = Parity[i & 0xF] ^ Parity[i >> 4];
    }

    Init = 1;
  }

  /* Opcode table has been build, reset CPU */
  POTATO_HardReset();
}

/*------------------------------------*/
/*-- Hard Reset the POTATO CPU core --*/
/*------------------------------------*/
/* Input : None                       */
/* Output: None                       */
/* Action: Reset CPU depend variables */
/*       : Clears RAM                 */
/*------------------------------------*/
/* Notes : None                       */
/*------------------------------------*/
void POTATO_HardReset()
{
  /* Clear RAM */
  memset(Ram[0], 0x00, 0x100);
  memset(Ram[1], 0x00, 0x100);

  /* Clear Work RAM */
  memset(WorkRam[0], 0x00, 0x100);
  memset(WorkRam[1], 0x00, 0x100);

  /* Clear LCD RAM */
  memset(XRam[0], 0x00, 0x80);
  memset(XRam[1], 0x00, 0x80);
  memset(XRam[2], 0x00, 0x80);

  Cpu.PC  = 0x0000;
  Cpu.SP  = 0x7F;
  Cpu.Acc = 0x00;
  Cpu.B   = 0x00;
  Cpu.C   = 0x00;
  Cpu.PSW = 0x00;
  SetPSW();
  Cpu.TRR.Value = 0x0000;
  Cpu.FPR  = 0x00;
  Cpu.PCON = 0x00;
  Cpu.IE   = 0x00;
  Cpu.IP   = 0x00;
  Cpu.EXT  = 0x01;
  Cpu.OCR  = 0x00;
  Cpu.MCR  = 0x00;
  Cpu.XBNK = 0x00;
  Cpu.VCCR = 0x00;

  Cpu.P1    = 0x00;
  Cpu.P3    = 0xFF; /* No Buttons Pressed */
  Cpu.P3INT = 0x00;
  Cpu.P7    = 0x02; /* Battery High, not inserted into game controller */

  Cpu.VSEL   = 0x00;
  Cpu.VRMAD1 = 0x00;
  Cpu.VRMAD2 = 0x00;

  Cpu.Timer0.T0CNT      = 0x00;
  Cpu.Timer0.T0PRR      = 0x00;
  Cpu.Timer0.Reload.T0R = 0x0000;
  Cpu.Timer0.Timer.T0   = 0x0000;
  Cpu.Timer0.PreScale   = 0;
  Cpu.Timer0.TimeBase   = 0;

  Cpu.BaseTimer.BTCR    = 0x00;
  Cpu.BaseTimer.Counter = 0x0000;

  if (BiosLoaded)
  {
    Cpu.Code = BiosRom;
  }
  else
  {
    Cpu.Code = FlashRom[0];
  }

  Cpu.Speed = 5000; /* (600Khz / 2) / 60 (RC Clock / 12) */
  Cpu.NrCycles = 0;
  Cpu.Halt     = 0;
  Cpu.Hold     = 0;
}

/*------------------------------------*/
/*--  Load VMS File into FlashRom   --*/
/*------------------------------------*/
/* Input : Filename                   */
/* Output: Error/Succes Code          */
/* Action: Load file into FlashRom    */
/*         bank0.                     */
/*------------------------------------*/
/* Notes : Filesize must be <= 64KB   */
/*------------------------------------*/
int POTATO_LoadVms(char *Name)
{
  FILE *hand;
  int  FileSize;

  /* Get handle to file */
  hand = fopen(Name, "rb");

  if (hand == NULL)
  {
    return 1;
  }

  /* Set file pointer to end of file */
  fseek(hand, 0, SEEK_END);
  
  /* Get file size */
  FileSize = ftell(hand);
  
  if (FileSize > 0x10000)
  {
    fclose(hand);
    return 1;
  }
  
  /* Set file pointer to start of file */
  fseek(hand, 0, SEEK_SET);
  
  /* Read in the data */
  fread(FlashRom[0], 1, FileSize, hand);
  
  /* Close file handle */
  fclose(hand);

  Cpu.EXT |= 1;
  Cpu.Code = FlashRom[0];
  
  return 0;
}

/*------------------------------------*/
/*--  Load BIOS File into FlashRom  --*/
/*------------------------------------*/
/* Input : Biosname                   */
/* Output: Error/Succes Code          */
/* Action: Load BIOS into BiosRom     */
/*------------------------------------*/
/* Notes : Filesize must be 64KB      */
/*------------------------------------*/
int POTATO_LoadBios(char *Name)
{
  FILE *hand;
  int  FileSize;

  /* Get handle to file */
  hand = fopen(Name, "rb");

  if (hand == NULL)
  {
    return 1;
  }

  /* Set file pointer to end of file */
  fseek(hand, 0, SEEK_END);
  
  /* Get file size */
  FileSize = ftell(hand);
  
  if (FileSize != 0x10000)
  {
    fclose(hand);
    return 1;
  }
  
  /* Set file pointer to start of file */
  fseek(hand, 0, SEEK_SET);
  
  /* Read in the data */
  fread(BiosRom, 1, FileSize, hand);
  
  /* Close file handle */
  fclose(hand);
  
  return 0;
}

/*------------------------------------*/
/*--      Press a (P3) Button       --*/
/*------------------------------------*/
/* Input : Button Nr (Bitwise)        */
/* Output:                            */
/*------------------------------------*/
/* Notes :                            */
/*------------------------------------*/
void POTATO_PressKey(int Key)
{
  Cpu.P3 &= ~(1 << Key);

  if (Cpu.P3INT & 0x4)
  {
    Cpu.P3INT |= 0x2;
  }

  Cpu.Halt = 0;

  if ((Cpu.P3INT & 0x3) == 0x3)
  {
    POTATO_Interrupt(10);
  }
}

/*------------------------------------*/
/*--     Release a (P3) Button      --*/
/*------------------------------------*/
/* Input : Button Nr (Bitwise)        */
/* Output:                            */
/*------------------------------------*/
/* Notes :                            */
/*------------------------------------*/
void POTATO_ReleaseKey(int Key)
{
  Cpu.P3 |= (1 << Key);

  if (Cpu.P3INT & 0x4)
  {
    Cpu.P3INT |= 0x2;
  }

  Cpu.Halt = 0;

  if ((Cpu.P3INT & 0x3) == 0x3)
  {    
    POTATO_Interrupt(10);
  }
}

/*------------------------------------*/
/*--     Generate an Interrupt      --*/
/*------------------------------------*/
/* Input : Interrupt Level            */
/* Output:                            */
/*------------------------------------*/
/* Notes :                            */
/*------------------------------------*/
void POTATO_Interrupt(int Level)
{
  if (IntMask == 3) return; /* Only 3 interrupts permitted */

  if (Cpu.IE & 0x80)
  {
    IntReq |= (1 << (Level - 1));
    IntMask++;

    Cpu.Halt = 0;
    Cpu.Hold = 0;

    if ((IntMask & 0x80) == 0)
      POTATO_ProcessInterrupt();
  }
}

/*------------------------------------*/
/*--        Process Interrupt       --*/
/*------------------------------------*/
/* Input :                            */
/* Output:                            */
/*------------------------------------*/
/* Notes :                            */
/*------------------------------------*/
void POTATO_ProcessInterrupt()
{
  int i;
  
  if (IntReq)
  {
    for (i = 0; i < 10; i++)
    {
      if (IntReq & (1 << i))
      {
        IntReq &= ~(1 << i);

        IntMask |= 0x80;
        Ram[0][++Cpu.SP] = Cpu.PC & 0xFF;
        Ram[0][++Cpu.SP] = Cpu.PC >> 8;

        Cpu.Halt = 0;
        Cpu.Hold = 0;
        Cpu.PC   = IntVector[i];
        
        return;
      }
    }
  }
}

void POTATO_RunTimers(int Cycles)
{
  int Ticks = 0;
  
  /* Timer 0 */
  if (Cpu.Timer0.T0CNT & 0xC0)
  {
    switch (Cpu.Timer0.T0CNT & 0x30)
    {
      case 0x00: /* Mode 0 */
        Cpu.Timer0.TimeBase += Cycles;

        while (Cpu.Timer0.TimeBase >= Cpu.Timer0.PreScale)
        {
          Cpu.Timer0.TimeBase -= Cpu.Timer0.PreScale;
          Ticks++;
        }

        if (Ticks)
        {
          /* High Timer */
          if (Cpu.Timer0.T0CNT & 0x80)
          {
            Cpu.Timer0.Timer.Byte.T0H += Ticks;
          
            /* Overflow */
            if (Cpu.Timer0.Timer.Byte.T0H == 0)
            {
              /* Set overflow flag */
              Cpu.Timer0.T0CNT |= 0x8;

              /* Reload High Data */
              Cpu.Timer0.Timer.Byte.T0H = Cpu.Timer0.Reload.Byte.T0HR;

              /* Generate interrupt */
              if (Cpu.Timer0.T0CNT & 0x4)
                POTATO_Interrupt(5);
            }
          }

          /* Low Timer */
          if (Cpu.Timer0.T0CNT & 0x40)
          {
            Cpu.Timer0.Timer.Byte.T0L += Ticks;
          
            /* Overflow */
            if (Cpu.Timer0.Timer.Byte.T0L == 0)
            {
              /* Set overflow flag */
              Cpu.Timer0.T0CNT |= 0x2;

              /* Reload High Data */
              Cpu.Timer0.Timer.Byte.T0L = Cpu.Timer0.Reload.Byte.T0LR;

              /* Generate interrupt */
              if (Cpu.Timer0.T0CNT & 0x1)
                POTATO_Interrupt(3);
            }
          }
        }
        break;

      case 0x10: /* Mode 1 */
        break;

      case 0x20: /* Mode 2 */
        Cpu.Timer0.TimeBase += Cycles;

        while (Cpu.Timer0.TimeBase >= Cpu.Timer0.PreScale)
        {
          Cpu.Timer0.TimeBase -= Cpu.Timer0.PreScale;
          Ticks++;
        }

        if (Ticks)
        {
          Cpu.Timer0.Timer.T0 += Ticks;

          if (Cpu.Timer0.Timer.T0 == 0)
          {
            /* Set overflow flag */
            Cpu.Timer0.T0CNT |= 0xA;

            /* Reload Data */
            Cpu.Timer0.Timer.T0 = Cpu.Timer0.Reload.T0R;

            /* Generate interrupt */
            if (Cpu.Timer0.T0CNT & 0x4)
              POTATO_Interrupt(5);
          }
        }
        break;

      case 0x30: /* Mode 3 */
        break;
    }
  }

  /* Base Timer */
  if (Cpu.BaseTimer.BTCR & 0x40)
  {
    if (Cpu.BaseTimer.BTCR & 0x1)
    {
      Cpu.BaseTimer.Counter += (Cycles << 1);

      /* 14-bit Counter Overflow */
      if (Cpu.BaseTimer.Counter & 0x4000)
      {
        Cpu.BaseTimer.Counter &= 0x3FFF;
        
        /* Set Interrupt Source */
        Cpu.BaseTimer.BTCR |= 0x2;

        /* Generate Interrupt */
        POTATO_Interrupt(4);
      }
    }
  }
}


void POTATO_Execute(int Instructions)
{
  int i;
  unsigned char Opcode;
  int Ticks;

  for (i = 0; i < Instructions; i++)
  {
    Opcode = Cpu.Code[Cpu.PC];
    Ticks = VMU_OpTable[Opcode](Opcode);
    POTATO_RunTimers(Ticks);
  }
}

void POTATO_ExecuteCycles(int Cycles)
{
  unsigned char Opcode;
  int Ticks;

  /* Hold mode */
  //if (Cpu.Hold) return;

  Cpu.NrCycles += Cycles;

  while (Cpu.NrCycles > 0)
  {
    /* Halt mode */
    if (Cpu.Halt)
    {
      POTATO_RunTimers(1);
      Cpu.NrCycles--;
    }
    else
    {
      Opcode = Cpu.Code[Cpu.PC];
      Ticks = VMU_OpTable[Opcode](Opcode);
      Cpu.NrCycles -= Ticks;
      POTATO_RunTimers(Ticks);
    }
  }
}