#include <windows.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include "resource.h"
#include "VMUCore.h"
#include "DreamVMU.h"

VOID UpdateCPURegs(HWND hDlg);
VOID UpdateCPUDisasm(HWND hDlg);
VOID RefreshMemoryViewer(HWND hDlg);
VOID CreateMemoryCombo(HWND hDlg);

int MemoryStart       = 0;
int MemoryEnd         = 0;
unsigned char *MemPtr = NULL;


LRESULT CALLBACK DebugProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	static UINT Timer;
  static UINT CPURunning;
  
  switch (message)
	{
		case WM_INITDIALOG:
    {
      UpdateCPURegs(hDlg);
      UpdateCPUDisasm(hDlg);
      CreateMemoryCombo(hDlg);
      RefreshMemoryViewer(hDlg);

      CPURunning = 0;
      return TRUE;
    }
		case WM_COMMAND:
      switch(HIWORD(wParam))
      {
        case CBN_SELCHANGE:
        {
          int Mem;
          MemoryStart = 0;
          Mem = SendMessage(GetDlgItem(hDlg, LOWORD(wParam)), CB_GETCURSEL, 0, 0);

          switch (Mem)
          {
            case 0:
              MemoryEnd = 0x100;
              MemPtr = Ram[0];
              break;
            case 1:
              MemoryEnd = 0x100;
              MemPtr = Ram[1];
              break;
            case 2:
              MemoryEnd = 0x100;
              MemPtr = WorkRam[0];
              break;
            case 3:
              MemoryEnd = 0x100;
              MemPtr = WorkRam[1];
              break;
            case 4:
              MemoryEnd = 0x80;
              MemPtr = XRam[0];
              break;
            case 5:
              MemoryEnd = 0x80;
              MemPtr = XRam[1];
              break;
            case 6:
              MemoryEnd = 0x80;
              MemPtr = XRam[2];
              break;
            case 7:
              MemoryEnd = 0x10000;
              MemPtr = FlashRom[0];
              break;
            case 8:
              MemoryEnd = 0x10000;
              MemPtr = FlashRom[1];
              break;
            case 9:
              MemoryEnd = 0x10000;
              MemPtr = BiosRom;
              break;
          }
  
          RefreshMemoryViewer(hDlg);
          break;
        }
        break;
      }

			switch (LOWORD(wParam))
      {
        case IDOK:
        case IDCANCEL:
        case IDC_POTATO_EXIT:
				  KillTimer(hDlg,Timer);
          EndDialog(hDlg, LOWORD(wParam));
				  return TRUE;

        case IDC_POTATO_RESET:
          if (CPURunning)
          {
            break;
          }
          
          /* Reset CPU */
          POTATO_HardReset();

          UpdateCPURegs(hDlg);
          UpdateCPUDisasm(hDlg);
          RefreshMemoryViewer(hDlg);
          break;

        case IDC_POTATO_STEP:
          if (CPURunning)
          {
            break;
          }
          
          /* Run 1 instruction */
          POTATO_Execute(1);

          UpdateCPURegs(hDlg);
          UpdateCPUDisasm(hDlg);
          RefreshMemoryViewer(hDlg);
          break;

        case IDC_POTATO_GO:
          if (CPURunning)
          {
            break;
          }

          /* Start timer interval */
          Timer = SetTimer(hDlg, 1, 2, NULL);
          CPURunning = 1;
          break;

        case IDC_POTATO_PAUSE:
          /* Kill the timer interval */
          KillTimer(hDlg,Timer);
          CPURunning = 0;
          break;

        case IDC_POTATO_SETPC:
         {
          int Index = SendMessage(GetDlgItem(hDlg, IDC_POTATO_DISASM_LIST), LB_GETCURSEL, 0, 0);

          int i;
          unsigned short j = Cpu.PC;

          for (i = 0; i < Index; i++)
          {
            j += POTATO_Disassemble(j, NULL);
          }
          
          Cpu.PC = j;

          UpdateCPURegs(hDlg);
          UpdateCPUDisasm(hDlg);
          break;
         }

        case IDC_MEMVIEWER_NEXT:
          if (MemoryStart + 0x80 < MemoryEnd)
          {
            MemoryStart += 0x80;
            RefreshMemoryViewer(hDlg);
          }
          break;

        case IDC_MEMVIEWER_PREV:
          if (MemoryStart - 0x80 >= 0)
          {
            MemoryStart -= 0x80;
            RefreshMemoryViewer(hDlg);
          }
          break;
			}
			break;
    
    case WM_TIMER:
      /* Run 1 instruction */
      POTATO_Execute(1);

      UpdateCPURegs(hDlg);
      UpdateCPUDisasm(hDlg);
      RefreshMemoryViewer(hDlg);
      break;
	}

	return FALSE;
}

VOID UpdateCPURegs(HWND hDlg)
{
  char temp[0x20];

  /* Registers */
  sprintf(temp, "0x%02x", Cpu.Acc);
  SetDlgItemText(hDlg, IDC_POTATO_ACC, temp);
  
  sprintf(temp, "0x%02x", Cpu.B);
  SetDlgItemText(hDlg, IDC_POTATO_B, temp);

  sprintf(temp, "0x%02x", Cpu.C);
  SetDlgItemText(hDlg, IDC_POTATO_C, temp);

  GetPSW();
  sprintf(temp, "0x%02x", Cpu.PSW);
  SetDlgItemText(hDlg, IDC_POTATO_PSW, temp);

  sprintf(temp, "0x%02x", Cpu.SP);
  SetDlgItemText(hDlg, IDC_POTATO_SP, temp);

  sprintf(temp, "0x%04x", Cpu.PC);
  SetDlgItemText(hDlg, IDC_POTATO_PC, temp);

  sprintf(temp, "0x%02x", Cpu.EXT);
  SetDlgItemText(hDlg, IDC_POTATO_EXT, temp);
  
  sprintf(temp, "0x%02x", Cpu.OCR);
  SetDlgItemText(hDlg, IDC_POTATO_OCR, temp);

  sprintf(temp, "0x%02x", Cpu.XBNK);
  SetDlgItemText(hDlg, IDC_POTATO_XBNK, temp);

  sprintf(temp, "0x%02x", Cpu.PCON);
  SetDlgItemText(hDlg, IDC_POTATO_PCON, temp);

  sprintf(temp, "0x%04x", Cpu.TRR.Value);
  SetDlgItemText(hDlg, IDC_POTATO_TRR, temp);

  sprintf(temp, "0x%04x", Cpu.Timer0.Timer.T0);
  SetDlgItemText(hDlg, IDC_POTATO_T0, temp);

  sprintf(temp, "0x%04x", Cpu.Timer0.Reload.T0R);
  SetDlgItemText(hDlg, IDC_POTATO_T0R, temp);

  sprintf(temp, "0x%02x", Cpu.Timer0.T0CNT);
  SetDlgItemText(hDlg, IDC_POTATO_T0CNT, temp);

  sprintf(temp, "0x%02x", Cpu.Timer0.T0PRR);
  SetDlgItemText(hDlg, IDC_POTATO_T0PRR, temp);

  sprintf(temp, "0x%02x", Cpu.BaseTimer.BTCR);
  SetDlgItemText(hDlg, IDC_POTATO_BTCR, temp);

  sprintf(temp, "0x%04x", Cpu.BaseTimer.Counter);
  SetDlgItemText(hDlg, IDC_POTATO_BASE, temp);

  /* Flags */
  sprintf(temp, "%d", Cpu.CY);
  SetDlgItemText(hDlg, IDC_POTATO_CY, temp);

  sprintf(temp, "%d", Cpu.AC);
  SetDlgItemText(hDlg, IDC_POTATO_AC, temp);

  sprintf(temp, "%d", Cpu.IRBK1);
  SetDlgItemText(hDlg, IDC_POTATO_IRBK1, temp);

  sprintf(temp, "%d", Cpu.IRBK0);
  SetDlgItemText(hDlg, IDC_POTATO_IRBK0, temp);

  sprintf(temp, "%d", Cpu.OV);
  SetDlgItemText(hDlg, IDC_POTATO_OV, temp);

  sprintf(temp, "%d", Cpu.RAMBK0);
  SetDlgItemText(hDlg, IDC_POTATO_RAMBK0, temp);

  Cpu.P = Parity[Cpu.Acc];
  sprintf(temp, "%d", Cpu.P);
  SetDlgItemText(hDlg, IDC_POTATO_P, temp);
}


/* -- Add Disassembled instructions to ListBox -- */
VOID UpdateCPUDisasm(HWND hDlg)
{
  HWND List;
  char Disasm[0x100];
  int i;
  int j = 0;

  /* Get Handle of ListBox */
  List = GetDlgItem(hDlg, IDC_POTATO_DISASM_LIST);
  
  /* Reset ListBox */
  SendMessage(List, LB_RESETCONTENT, 0, 0);
	
  /* Add Strings to ListBox */
  for (i = 0;i < 20;i++)
  {
    j += POTATO_Disassemble(Cpu.PC + j, Disasm);

    SendMessage(List, LB_ADDSTRING, 0,(long) Disasm);
  }
  
  /* Set current selected position */
  SendMessage(List,LB_SETCURSEL,0,0);
}

VOID RefreshMemoryViewer(HWND hDlg)
{
  char temp[10];
  int i;

  /* Fill in address values */
  for (i = 0; i < 8; i++)
  {
    sprintf(temp,"0x%04x", MemoryStart + (i * 0x10));
    SetDlgItemText(hDlg, IDC_MEMVIEWER_LABEL_0 + i,temp);
  }

  /* Fill in data values */
  for (i = 0; i < 128; i++)
  {
    if ((MemoryStart + i) > MemoryEnd)
    {
      SetDlgItemText(hDlg, IDC_MEMVIEWER_0 + i, (LPCSTR) "");
    }
    else
    {
      sprintf(temp, "%02x", MemPtr[MemoryStart + i]);
      SetDlgItemText(hDlg, IDC_MEMVIEWER_0 + i, CharUpper((LPTSTR) temp));
    }
  }
}

VOID CreateMemoryCombo(HWND hDlg)
{
  HWND Combo;
  Combo = GetDlgItem(hDlg, IDC_MEMVIEWER_COMBO);
  
  SendMessage(Combo, CB_RESETCONTENT, 0, 0);

	SendMessage(Combo, CB_ADDSTRING, 0, (long) "Ram Bank0");
  SendMessage(Combo, CB_ADDSTRING, 0, (long) "Ram Bank1");
  SendMessage(Combo, CB_ADDSTRING, 0, (long) "WorkRam Bank0");
  SendMessage(Combo, CB_ADDSTRING, 0, (long) "WorkRam Bank1");
  SendMessage(Combo, CB_ADDSTRING, 0, (long) "XRam Bank0");
  SendMessage(Combo, CB_ADDSTRING, 0, (long) "XRam Bank1");
  SendMessage(Combo, CB_ADDSTRING, 0, (long) "XRam Bank2");
  SendMessage(Combo, CB_ADDSTRING, 0, (long) "FlashRom Bank0");
  SendMessage(Combo, CB_ADDSTRING, 0, (long) "FlashRom Bank1");
  SendMessage(Combo, CB_ADDSTRING, 0, (long) "BiosRom");
  
  SendMessage(Combo, CB_SETCURSEL, 0, 0);

  MemoryStart = 0;
  MemoryEnd   = 0x100;
  MemPtr      = &Ram[0][0];
}