#include <windows.h>
#include <stdio.h>
#include "DreamVMU.h"
#include "DirectX.h"

LPDIRECTDRAW7        lpDD;
LPDIRECTDRAWSURFACE7 lpDDS_Prim;
LPDIRECTDRAWSURFACE7 lpDDS_Back;
LPDIRECTDRAWCLIPPER  lpDDC_Clip;
RECT                 ScreenRect;

unsigned long        FGColor = 0xFFFFFF;
unsigned long        BGColor = 0x000000;

int DirectDraw_Init(HWND hWnd)
{
	DDSURFACEDESC2 ddsd;

	DirectDraw_Exit();
	
	/* Create DirectDraw Interface */
  if (DD_OK != DirectDrawCreateEx(NULL, (LPVOID *) &lpDD, IID_IDirectDraw7, NULL))
  {
    /* Failed */
    ShowMessage("DirectDraw: DirectDrawCreateEx failed");
    return 1;
  }

	/* Set cooperative levels */
  if (DD_OK != lpDD->SetCooperativeLevel(hWnd, DDSCL_NORMAL))
  {
    ShowMessage("DirectDraw: SetCooperativeLevel in windowed mode failed");
    return 1;
  }
	
	/* Create primary surface */
  memset(&ddsd, 0, sizeof(ddsd));

  ddsd.dwSize         = sizeof(ddsd);
	ddsd.dwFlags        = DDSD_CAPS;
	ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;

	if (DD_OK != lpDD->CreateSurface(&ddsd, &lpDDS_Prim, NULL))
  {
		ShowMessage("DirectDraw: Primary surface creation failed");
    return 1;
  }

	/* Create clipper */
  if (DD_OK != lpDD->CreateClipper(0, &lpDDC_Clip, NULL ))
  {
    ShowMessage("DirectDraw: CreateClipper failed");
    return 1;
  }

	if (DD_OK != lpDDC_Clip->SetHWnd(0, hWnd))
  {
		ShowMessage("DirectDraw: SetHWnd failed");
    return 1;
  }

	if (DD_OK != lpDDS_Prim->SetClipper(lpDDC_Clip))
  {
		ShowMessage("DirectDraw: SetClipper failed");
    return 1;
  }

  lpDDC_Clip->Release();
  lpDDC_Clip = NULL;

  memset(&ddsd, 0, sizeof(ddsd));
	ddsd.dwSize         = sizeof(ddsd);
	ddsd.dwFlags        = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH;
	ddsd.ddsCaps.dwCaps = DDSCAPS_SYSTEMMEMORY;
	ddsd.dwWidth        = ResX;
	ddsd.dwHeight       = ResY;

	if (DD_OK != lpDD->CreateSurface(&ddsd, &lpDDS_Back, NULL))
  {
		ShowMessage("DirectDraw: Back surface creation failed");
    return 1;
  }

	return 0;
}

void DirectDraw_Exit()
{
	if (lpDDS_Back)
	{
		lpDDS_Back->Release();
		lpDDS_Back = NULL;
	}

	if (lpDDS_Prim)
	{
		lpDDS_Prim->Release();
		lpDDS_Prim = NULL;
	}

  if (lpDDC_Clip)
	{
		lpDDC_Clip->Release();
		lpDDC_Clip = NULL;
	}

	if (lpDD)
	{
		lpDD->SetCooperativeLevel(DreamHWnd, DDSCL_NORMAL);
		lpDD->Release();
		lpDD = NULL;
	}
}

void DirectDraw_ReSize(HWND hWnd)
{
  GetWindowRect(hWnd, &ScreenRect);

	ScreenRect.left   += 56; 
	ScreenRect.top    += 169;
	ScreenRect.right  = ScreenRect.left + ResX;//-= 62;
	ScreenRect.bottom = ScreenRect.top + ResY;//-= 178;
}

void DirectDraw_ClearBackSurface()
{
	DDBLTFX ddbltfx;
	
  memset(&ddbltfx, 0, sizeof(ddbltfx));

	ddbltfx.dwSize      = sizeof(ddbltfx);
	ddbltfx.dwFillColor = BGColor;
	
  lpDDS_Back->Blt(NULL, NULL, NULL, DDBLT_COLORFILL | DDBLT_WAIT, &ddbltfx);
}

void DirectDraw_Paint(HWND hWnd)
{
	HRESULT rval;
	
	rval = lpDDS_Prim->Blt(&ScreenRect, lpDDS_Back, NULL, DDBLT_WAIT, NULL);

	if (rval == DDERR_SURFACELOST)
  {
    lpDDS_Prim->Restore();
    lpDDS_Back->Restore();
  }
}