#include <windows.h>
#include <stdio.h>
#include "DreamVMU.h"
#include "DirectX.h"

LPDIRECTSOUND8          lpDirectSound;
LPDIRECTSOUNDBUFFER8    lpSoundBuffer8;
LPDIRECTSOUNDBUFFER     lpSoundBuffer;

int SoundEnabled = 1;
int SoundRate    = 44100;
int SoundStereo  = 1;
int SoundSegs    = 8;
int SoundSize;
int WritePtr;
int ReadPtr;
int BytesPerUnit;

int Sound[2][735];

/* -- Initialise DirectSound -- */
int DirectSound_Init(HWND hWnd)
{
  /* Create DirectSound object, and set Cooperative level */
  if (DS_OK == DirectSoundCreate8(NULL,&lpDirectSound,NULL))
  {
    if (DS_OK != lpDirectSound->SetCooperativeLevel(hWnd,DSSCL_EXCLUSIVE))
    {
      if (DS_OK != lpDirectSound->SetCooperativeLevel(hWnd,DSSCL_NORMAL))
      {
        /* Release DirectSound object */
        lpDirectSound->Release();
        lpDirectSound = NULL;
        
        SoundEnabled = 0;
        return 1;
      }
    }
  }
  else 
  {
    SoundEnabled = 0; 
    return 1;
  }

  ReadPtr  = 0;
  WritePtr = 0;

  WAVEFORMATEX wfx;
  memset(&wfx, 0, sizeof(WAVEFORMATEX));

  wfx.wFormatTag = WAVE_FORMAT_PCM;

  /* Check the sound rate */
  switch (SoundRate)
  {
     case 11025:
        wfx.nSamplesPerSec = 11025;
        SoundSize = 184;
        break;
     
     case 22050:
        wfx.nSamplesPerSec = 22050;
        SoundSize = 368;
        break;
     
     case 44100:
        wfx.nSamplesPerSec = 44100;
        SoundSize = 735;
        break;
     
     default:
        wfx.nSamplesPerSec = 44100;
        SoundSize = 735;
        break;
  }

  /* Check if sound is stereo */
  if (SoundStereo )
  {
     wfx.nChannels   = 2;
  }
  else
  {
     wfx.nChannels   = 1;
  }

  wfx.wBitsPerSample  = 16;
  wfx.nBlockAlign     = BytesPerUnit = wfx.nChannels * (wfx.wBitsPerSample / 8); 
  wfx.nAvgBytesPerSec = wfx.nSamplesPerSec * BytesPerUnit;
  wfx.cbSize          = sizeof(WAVEFORMATEX);
    
  
  DSBUFFERDESC dsbd;
  memset(&dsbd, 0, sizeof(DSBUFFERDESC));
 
  dsbd.dwSize          = sizeof(DSBUFFERDESC);
  dsbd.dwFlags         = DSBCAPS_STICKYFOCUS | DSBCAPS_CTRLVOLUME;
  dsbd.dwBufferBytes   = SoundSize * SoundSegs * BytesPerUnit; 
  dsbd.lpwfxFormat     = &wfx;
  dsbd.guid3DAlgorithm = GUID_NULL;
    
  /* Create sound buffer */
  if (DS_OK != lpDirectSound->CreateSoundBuffer(&dsbd, &lpSoundBuffer, NULL))
  {
    SoundEnabled = 0; 
    return 1;
  }
  
  /* Get DirectSoundBuffer8 object */
  if (DS_OK != lpSoundBuffer->QueryInterface(IID_IDirectSoundBuffer8, (LPVOID *) &lpSoundBuffer8))
  {
    SoundEnabled = 0; 
    return 1;
  }

  lpSoundBuffer->Release();
  lpSoundBuffer = NULL;
  
  /* Play the buffer */
  if (DS_OK != lpSoundBuffer8->Play(0, 0, DSBPLAY_LOOPING))
  {
    SoundEnabled = 0; 
    return 1;
  }
       
  SoundEnabled = 1; 
  return 0;
}

/* -- ShutDown DirectSound -- */
void DirectSound_Exit()
{
  if (SoundEnabled)
  {
    if (lpSoundBuffer)
    {
      lpSoundBuffer->Release();
      lpSoundBuffer = NULL;
    }

    if (lpSoundBuffer8)
    {
      lpSoundBuffer8->Stop();
      lpSoundBuffer8->Release();
      lpSoundBuffer8 = NULL;
    }

    if (lpDirectSound)
    {
      lpDirectSound->Release();
      lpDirectSound = NULL;
    }

    SoundEnabled = 0;
  }
}

/* -- Clear the SoundBuffer -- */
int DirectSound_Silence()
{
  LPVOID lpPtr1;
	DWORD  dwSize1;
  DWORD  TotalSize;
  HRESULT ret;

  /* Check if DirectSound is initialized correctly */
  if (SoundEnabled == 0)
    return 1;
	
	TotalSize = SoundSegs * SoundSize * BytesPerUnit;
  
	ret = lpSoundBuffer8->Lock(0, TotalSize, &lpPtr1,&dwSize1, NULL, NULL, 0);
	
	if (ret == DS_OK) 
  {
		memset(lpPtr1, 0, TotalSize);
		lpSoundBuffer8->Unlock(lpPtr1, dwSize1, NULL, NULL);
	}
	
  return 0;
}

/* -- Write new data to SoundBuffer -- */
int DirectSound_Write()
{
  LPVOID lpPtr1;
	DWORD  dwSize1;
  DWORD  dwOffset;
  DWORD  dwBytes;
  HRESULT ret;
  
  /* Check if DirectSound is initialized correctly */
  if (SoundEnabled == 0) 
    return 1;

  dwOffset = WritePtr * SoundSize * BytesPerUnit; 
  dwBytes  = SoundSize * BytesPerUnit; 

  /* Lock the soundbuffer */
  ret = lpSoundBuffer8->Lock(dwOffset,dwBytes,&lpPtr1,&dwSize1,NULL,NULL,0);
  
  /* Check if soundbuffer is lost */
  if (ret == DSERR_BUFFERLOST)
  {
    lpSoundBuffer8->Restore();
    ret = lpSoundBuffer8->Lock(dwOffset,dwBytes,&lpPtr1,&dwSize1,NULL,NULL,0);

    if (ret != DS_OK)
      return 1;
  }

  if (ret == DS_OK)
  {
    int i;
    
    signed short *convertSound16;

    if (SoundStereo )
    {
      convertSound16 = (signed short *) lpPtr1;
      int L, R;
      for (i = 0;i < SoundSize;i++)
      {
        //convertSound16[i * 2    ] = Sound[0][i];
        //convertSound16[i * 2 + 1] = Sound[1][i];

        L = Sound[0][i];
        if (L < -0x7FFF) *convertSound16++ = -0x7FFF;
        else if (L > 0x7FFF) *convertSound16++ = 0x7FFF;
        else *convertSound16++ = (short) L;

        R = Sound[1][i];
        if (R < -0x7FFF) *convertSound16++ = -0x7FFF;
        else if (R > 0x7FFF) *convertSound16++ = 0x7FFF;
        else *convertSound16++ = (short) R;
      }
    }
    else
    {
      convertSound16 = (signed short *) lpPtr1;
      for (i = 0;i < SoundSize;i++)
      {
        convertSound16[i] = (Sound[0][i] >> 1) + (Sound[1][i] >> 1);
      }
    }
        
    lpSoundBuffer8->Unlock(lpPtr1,dwSize1,NULL,NULL);
  }

  return 0;
}

/* -- Get Current Play Segment -- */
int DirectSound_GetCurrentSeg()
{
  DWORD posPlay;
  lpSoundBuffer8->GetCurrentPosition(&posPlay,NULL);

  return (posPlay / (SoundSize * BytesPerUnit)); 
}