#include <windows.h>
#include <commctrl.h>
#include <stdio.h>
#include <string.h>
#include "resource.h"
#include "DreamVMU.h"
#include "VMUCore.h"
#include "DirectX.h"

HINSTANCE hInstance;
HWND      DreamHWnd;
HMENU     DreamMenu;
WNDCLASS  WndClass;
HACCEL    hAccelTable;
HDC       SkinDevice;
HRGN      ScreenRegion;

DWORD     SkinResource = IDB_VMUSKIN_1;

int DreamVMU_Running = 0;
int DreamVMU_Loaded  = 0;

void UpdatePOTATO();

/*-------------*/
/* Entry point */
/*-------------*/
int WINAPI WinMain(HINSTANCE hInst,	HINSTANCE hPrevInst, LPSTR lpCmdLine, int nCmdShow)
{
  MSG msg;

  /* Init everything */
  if (!DreamVMUInit(hInst, nCmdShow))
  {
    /* Display error message */
    MessageBox(NULL, "DreamVMU couldn't initialize correctly,\nand will now shut down.", "DreamVMU v"VMU_VERSION" Fatal Eror", MB_OK | MB_ICONSTOP);
    return 0;
  }

  /* Enter Message loop */
  while (DreamVMU_Running)
  {
    while (PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE))
    {
      if (!GetMessage(&msg, NULL, 0, 0)) return msg.wParam;
      if (!TranslateAccelerator (msg.hwnd, hAccelTable, &msg)) 
      {
		    TranslateMessage(&msg);
			  DispatchMessage(&msg);
		  }
    }
    
    if (DreamVMU_Loaded)
    {
      UpdatePOTATO();
    }

	}

  /* Shutdown everything */
  DreamVMUExit();

  /* Process Remaining Messages */
  while (PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE))
  {
    if (!GetMessage(&msg, NULL, 0, 0)) return msg.wParam;
  }

  return 0;
}

/*------------------------*/
/* Initializes everything */
/*------------------------*/
BOOL DreamVMUInit(HINSTANCE hInst, int nCmdShow)
{  
  /* Register DreamVMU */
  WndClass.style         = CS_OWNDC;
	WndClass.lpfnWndProc   = WndProc;
	WndClass.cbClsExtra    = 0;
	WndClass.cbWndExtra    = 0;
	WndClass.hInstance     = hInst;
	WndClass.hIcon         = LoadIcon(hInst, MAKEINTRESOURCE(IDI_ICON));
	WndClass.hCursor       = LoadCursor(NULL, IDC_ARROW);
	WndClass.hbrBackground = (HBRUSH)(COLOR_WINDOW + 2);
	WndClass.lpszMenuName  = NULL;
	WndClass.lpszClassName = "DreamVMU";

  if (!RegisterClass(&WndClass))
  {
    return FALSE;
  }

  hInstance = hInst;

  /* Create Window Handle */
  DreamHWnd = CreateWindow(
      "DreamVMU",
      "DreamVMU v"VMU_VERSION,
      WS_OVERLAPPEDWINDOW,
      CW_USEDEFAULT,
      CW_USEDEFAULT,
      304,
      534,
      NULL,
      NULL,
      hInst,
      NULL);

  if (DreamHWnd == NULL)
  {
    return FALSE;
  }

  /* Create menu */
  GenerateMenu();

  /* Remove outline border, and systemicon box */
  DWORD dwStyle = GetWindowLong(DreamHWnd, GWL_STYLE);
  dwStyle &= ~(WS_CAPTION | WS_SIZEBOX);
  SetWindowLong(DreamHWnd, GWL_STYLE, dwStyle);

  /* Get bimap handle from resource */
  HBITMAP SkinBitmap = LoadBitmap(hInst, MAKEINTRESOURCE(SkinResource));
  if (!SkinBitmap) return FALSE;

  /* Scan bitmap to get window region */
  HRGN SkinRgn = ScanRegion(SkinBitmap, 255, 0, 0);

  /* Create a device context for the skin */
  SkinDevice = CreateCompatibleDC(0);
  SelectObject(SkinDevice, SkinBitmap);
  DeleteObject(SkinBitmap);
  
  /* Set new created window region */
  SetWindowRgn(DreamHWnd, SkinRgn, true);

  /* Initialize DirectDraw */
  if (DirectDraw_Init(DreamHWnd))
  {
    ShowMessage("Error with DirectDraw");
    return FALSE;
  }

  /* Initialize DirectSound */
  if (DirectSound_Init(DreamHWnd))
  {
    ShowMessage("Error with DirectSound");
    return FALSE;
  }

  /* Initialize emulation core */
  POTATO_Init();
  
  /* Show window */
  ShowWindow(DreamHWnd, nCmdShow);

  DreamVMU_Running = 1;

  return TRUE;
}

/*-----------------*/
/* Exit everything */
/*-----------------*/
VOID DreamVMUExit()
{
  /* Release the Skin DC */
  DeleteDC(SkinDevice);

  DirectSound_Exit();
  DirectDraw_Exit();

  /* Destroy Main window */
  DestroyWindow(DreamHWnd);
}

/*----------------------*/
/* Main window Callback */
/*----------------------*/
LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{    
  switch (message) 
	{
    case WM_COMMAND:
			switch (LOWORD(wParam)) /* Parse menu, or accelarators */
			{
        case IDM_FILE_EXIT:
          DreamVMU_Running = 0;
          break;

        case IDM_LOAD_VMS:
          if (GetVmsName() == 0)
          {
            DreamVMU_Loaded = 1;
          }
          break;

        case IDM_LOAD_BIOS:
          if (GetBiosName() == 0)
          {
            DreamVMU_Loaded = 1;
          }
          break;

        case IDM_SKINS_1:
        {
          SkinResource = IDB_VMUSKIN_1;

          /* Create new bitmap handle */
          HBITMAP SkinBitmap = LoadBitmap(hInstance, MAKEINTRESOURCE(SkinResource));
          
          /* Create a device context for the skin */
          SkinDevice = CreateCompatibleDC(0);
          SelectObject(SkinDevice, SkinBitmap); 
          DeleteObject(SkinBitmap);

          /* Force window repaint */
          InvalidateRect(DreamHWnd, NULL, TRUE);
          SetWindowPos(DreamHWnd, NULL, 0, 0, 304, 534, SWP_NOMOVE| SWP_NOZORDER);

          /* Rebuild menu */
          GenerateMenu();

          break;
        }

        case IDM_SKINS_2:
        {
          SkinResource = IDB_VMUSKIN_2;

          /* Create new bitmap handle */
          HBITMAP SkinBitmap = LoadBitmap(hInstance, MAKEINTRESOURCE(SkinResource));
          
          /* Create a device context for the skin */
          SkinDevice = CreateCompatibleDC(0);
          SelectObject(SkinDevice, SkinBitmap); 
          DeleteObject(SkinBitmap);

          /* Force window repaint */
          InvalidateRect(DreamHWnd, NULL, TRUE);
          SetWindowPos(DreamHWnd, NULL, 0, 0, 304, 534, SWP_NOMOVE | SWP_NOZORDER);

          /* Rebuild menu */
          GenerateMenu();

          break;
        }

        case IDM_OPTIONS_DEBUG:
          DialogBox(hInstance, (LPCTSTR)IDD_DEBUG_DISASM, hWnd, (DLGPROC)DebugProc);
          break;

        case IDM_COLOR_FG:
        {
          CHOOSECOLOR CColor;
          COLORREF Colors[16];

          unsigned char R, G, B;

          memset(&CColor, 0, sizeof(CColor));
          memset(Colors, 0, sizeof(Colors));

          CColor.lStructSize  = sizeof(CColor);
          CColor.hwndOwner    = DreamHWnd;
          CColor.rgbResult    = 0x00FFFFFF;
          CColor.lpCustColors = Colors;
          CColor.Flags        = CC_RGBINIT | CC_SHOWHELP;

          if (ChooseColor(&CColor) == TRUE)
          {
            R = (GetRValue(CColor.rgbResult) >> 3);
            G = (GetGValue(CColor.rgbResult) >> 2);
            B = (GetBValue(CColor.rgbResult) >> 3);

            FGColor = (R << 11) | (G << 5) | B;

            POTATO_DrawLCD();
            DirectDraw_Paint(DreamHWnd);
          }

          break;
        }

        case IDM_COLOR_BG:
        {
          CHOOSECOLOR CColor;
          COLORREF Colors[16];

          unsigned char R, G, B;

          memset(&CColor, 0, sizeof(CColor));
          memset(Colors, 0, sizeof(Colors));

          CColor.lStructSize  = sizeof(CColor);
          CColor.hwndOwner    = DreamHWnd;
          CColor.rgbResult    = 0x00000000;
          CColor.lpCustColors = Colors;
          CColor.Flags        = CC_RGBINIT | CC_SHOWHELP;

          if (ChooseColor(&CColor) == TRUE)
          {
            R = (GetRValue(CColor.rgbResult) >> 3);
            G = (GetGValue(CColor.rgbResult) >> 2);
            B = (GetBValue(CColor.rgbResult) >> 3);

           BGColor = (R << 11) | (G << 5) | B;

            POTATO_DrawLCD();
            DirectDraw_Paint(DreamHWnd);
          }

          break;
        }
        
        case IDM_HELP_ABOUT:
          DialogBox(hInstance, (LPCTSTR)IDD_ABOUT, hWnd, (DLGPROC)DlgProc);
          break;

        default: /* Not handled commands */
				  return DefWindowProc(hWnd, message, wParam, lParam);
			}
			break;

    case WM_CLOSE: /* X Clicked */
      DreamVMU_Running = 0;
      break;

    case WM_MOVE:
      DirectDraw_ReSize(DreamHWnd);
      break;

    case WM_SIZE:
      DirectDraw_ReSize(DreamHWnd);
      break;

    case WM_PAINT:
      PAINTSTRUCT ps;
		  
      BeginPaint(DreamHWnd, &ps);

      BitBlt(ps.hdc, 0, 0, 307, 534, SkinDevice, 0, 0, SRCCOPY);

      DirectDraw_Paint(DreamHWnd);      
			
      EndPaint(DreamHWnd, &ps);
      break;

    case WM_RBUTTONDOWN: /* Right mouse button down */
    {
      POINT CurPos;

      GetCursorPos(&CurPos);

      TrackPopupMenu(DreamMenu, TPM_LEFTALIGN | TPM_TOPALIGN, CurPos.x, CurPos.y, NULL, hWnd, NULL);
      break;
    }

    case WM_LBUTTONDOWN: /* Left mouse button down */
    {
      int XPos = LOWORD(lParam);
      int YPos = HIWORD(lParam);

      if ((XPos >= 284) && (XPos <= 296) && (YPos >= 28) && (YPos <= 41))
      {
        /* Close button clicked */
        SendMessage(DreamHWnd, WM_CLOSE, NULL, NULL);
      }
      else if ((XPos >= 268) && (XPos <= 280) && (YPos >= 28) && (YPos <= 41))
      {
        /* Minimize button clicked */
        ShowWindow(DreamHWnd, SW_MINIMIZE);
      }
      else if ((XPos >= 12) && (XPos <= 24) && (YPos >= 28) && (YPos <= 41))
      {
        /* Menu button clicked */
        POINT CurPos;

        GetCursorPos(&CurPos);

        TrackPopupMenu(DreamMenu, TPM_LEFTALIGN | TPM_TOPALIGN, CurPos.x, CurPos.y, NULL, hWnd, NULL);
      }
      else
      {
        /* Move window */
        SendMessage(DreamHWnd, WM_NCLBUTTONDOWN, HTCAPTION, NULL);
      }
      break;
    }

    case WM_KEYDOWN:
			switch(wParam & 0xff) {
			case VK_UP:		 POTATO_PressKey(0);	 return 0;
			case VK_DOWN:	 POTATO_PressKey(1);	 return 0;
			case VK_LEFT:	 POTATO_PressKey(2);	 return 0;
			case VK_RIGHT: POTATO_PressKey(3); return 0;
			case 'Z':
			case 'A':		POTATO_PressKey(4); return 0;
			case 'X':
			case 'B':		POTATO_PressKey(5); return 0;
			case VK_RETURN:
			case 'M':		POTATO_PressKey(6); return 0;
			case VK_SPACE:
			case 'S':		POTATO_PressKey(7); return 0;
			}
			break;
		case WM_KEYUP:
			switch(wParam & 0xff) {
			case VK_UP:		 POTATO_ReleaseKey(0);	 return 0;
			case VK_DOWN:	 POTATO_ReleaseKey(1);	 return 0;
			case VK_LEFT:	 POTATO_ReleaseKey(2);	 return 0;
			case VK_RIGHT: POTATO_ReleaseKey(3); return 0;
			case 'Z':
			case 'A':		POTATO_ReleaseKey(4); return 0;
			case 'X':
			case 'B':		POTATO_ReleaseKey(5); return 0;
			case VK_RETURN:
			case 'M':		POTATO_ReleaseKey(6); return 0;
			case VK_SPACE:
			case 'S':		POTATO_ReleaseKey(7); return 0;
			}
			break;

		default:
			return DefWindowProc(hWnd, message, wParam, lParam);
   }

	return 0;
}

/*-----------------------*/
/* About dialog Callback */
/*-----------------------*/
LRESULT CALLBACK DlgProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch (message)
	{
		case WM_INITDIALOG:
      return TRUE;

		case WM_COMMAND:
			if ((LOWORD(wParam) == IDOK) || (LOWORD(wParam) == IDCANCEL))
			{
				EndDialog(hDlg, LOWORD(wParam));
				return TRUE;
			}
			break;
	}
    return FALSE;
}

/*-------------------*/
/* Display a message */
/*-------------------*/
VOID ShowMessage(const char *Text, ...)
{
  char temp[MAX_PATH];
  
  /* Check all parameters */
  va_list arg;
  va_start(arg,Text);

  vsprintf (temp,Text,arg);
  va_end(arg);

  /* Print message on screen */
  MessageBox(DreamHWnd, temp, "DreamVMU v"VMU_VERSION" Message...", MB_OK );
}

void UpdatePOTATO()
{
  while (WritePtr == DirectSound_GetCurrentSeg());

  ReadPtr = DirectSound_GetCurrentSeg();

  while (WritePtr != ReadPtr)
  {
    /* Write sound to sound buffer */
    DirectSound_Write();

    WritePtr = (WritePtr + 1) & (SoundSegs - 1);

    if (WritePtr != ReadPtr)
    {
      POTATO_ExecuteCycles(Cpu.Speed);
    }
    else
    {        
      POTATO_ExecuteCycles(Cpu.Speed);
      POTATO_DrawLCD();
      DirectDraw_Paint(DreamHWnd);
    }
  }
}